""" test DataDescription """

import datetime
import unittest

from aind_data_schema.data_description import (
    DataDescription,
    DerivedDataDescription,
    RawDataDescription,
    Funding,
)


class DataDescriptionTest(unittest.TestCase):
    """test DataDescription"""

    BAD_NAME = "fizzbuzz"
    BASIC_NAME = "ecephys_1234_3033-12-21_04-22-11"
    DERIVED_NAME = (
        "ecephys_1234_3033-12-21_04-22-11_spikesorted-ks25_2022-10-12_23-23-11"
    )

    def test_from_name(self):
        """test the from_name methods"""

        f = Funding(funder="test")

        da = DataDescription.from_name(
            name=self.BASIC_NAME,
            institution="AIND",
            data_level="raw data",
            funding_source=[f],
            modality="exaSPIM",
            subject_id="12345",
        )
        assert da.name == self.BASIC_NAME

        with self.assertRaises(ValueError):
            DataDescription.from_name(
                name=self.BAD_NAME,
                institution="AIND",
                data_level="raw data",
                funding_source=[f],
            )

        rd = RawDataDescription.from_name(
            name=self.BASIC_NAME, institution="AIND", funding_source=[f]
        )
        assert rd.name == self.BASIC_NAME
        assert rd.data_level.value == "raw data"

        with self.assertRaises(ValueError):
            RawDataDescription.from_name(
                name=self.BAD_NAME,
                institution="AIND",
                data_level="raw data",
                funding_source=[f],
            )

        dd = DerivedDataDescription.from_name(
            name=self.DERIVED_NAME,
            institution="AIND",
            funding_source=[f],
            modality="SmartSPIM",
            subject_id="12345",
        )
        assert dd.name == self.DERIVED_NAME
        assert dd.data_level.value == "derived data"

        with self.assertRaises(ValueError):
            DerivedDataDescription.from_name(
                name=self.BAD_NAME,
                institution="AIND",
                data_level="raw data",
                funding_source=[f],
            )

    def test_from_data_description(self):
        """test the from_data_description method"""
        dt = datetime.datetime.now()
        d1 = RawDataDescription(
            creation_date=dt.date(),
            creation_time=dt.time(),
            institution="AIND",
            data_level="raw data",
            funding_source=[],
            modality="ecephys",
            subject_id="12345",
        )

        dt = datetime.datetime.now()
        d2 = DerivedDataDescription.from_data_description(
            input_data=d1,
            process_name="fishing",
            creation_date=dt.date(),
            creation_time=dt.time(),
            institution="AIND",
            funding_source=[],
        )

        assert d2.modality == d1.modality
        assert d2.subject_id == d1.subject_id

        d3 = DerivedDataDescription.from_data_description(
            input_data=d2,
            process_name="bailing",
            creation_date=dt.date(),
            creation_time=dt.time(),
            institution="HUST",
            funding_source=[],
        )

        assert d3.modality == d2.modality
        assert d3.subject_id == d2.subject_id

    def test_constructors(self):
        """test building from component parts"""
        f = Funding(funder="test")

        dt = datetime.datetime.now()
        da = RawDataDescription(
            creation_date=dt.date(),
            creation_time=dt.time(),
            institution="AIND",
            data_level="raw data",
            funding_source=[f],
            modality="ecephys",
            subject_id="12345",
        )

        r1 = DerivedDataDescription(
            input_data_name=da.name,
            process_name="spikesort-ks25",
            creation_date=dt.date(),
            creation_time=dt.time(),
            institution="AIND",
            funding_source=[f],
            modality=da.modality,
            subject_id=da.subject_id,
        )

        r2 = DerivedDataDescription(
            input_data_name=r1.name,
            process_name="some-model",
            creation_date=dt.date(),
            creation_time=dt.time(),
            institution="AIND",
            funding_source=[f],
            modality="ecephys",
            subject_id="12345",
        )

        r3 = DerivedDataDescription(
            input_data_name=r2.name,
            process_name="a-paper",
            creation_date=dt.date(),
            creation_time=dt.time(),
            institution="AIND",
            funding_source=[f],
            modality="ecephys",
            subject_id="12345",
        )
        assert r3 is not None

        dd = DataDescription(
            label="test_data",
            modality="ecephys",
            subject_id="1234",
            data_level="raw data",
            creation_date=dt.date(),
            creation_time=dt.time(),
            institution="AIND",
            funding_source=[f],
        )

        assert dd is not None


if __name__ == "__main__":
    unittest.main()



""" Schemas for Ophys Sessions """

from __future__ import annotations

from datetime import datetime
from enum import Enum
from typing import List, Optional

from pydantic import Field
from ..base import AindCoreModel, AindModel


class FiberName(Enum):
    """Fiber name"""

    LASER_A = "Laser A"
    LASER_B = "Laser B"
    LASER_C = "Laser C"
    LASER_D = "Laser D"
    LASER_E = "Laser E"


class PatchCordName(Enum):
    """Patch cord name"""

    PATCH_CORD_A = "Patch Cord A"
    PATCH_CORD_B = "Patch Cord B"
    PATCH_CORD_C = "Patch Cord C"


class Coupling(AindModel):
    """Description of fiber coupling"""

    fiber_name: FiberName = Field(..., title="Fiber name")
    patch_cord_name: PatchCordName = Field(..., title="Patch cord name")


class Detector(AindModel):
    """Description of detector"""

    name: str = Field(..., title="Name")
    exposure_time: float = Field(..., title="Exposure time (ms)")


class LaserName(Enum):
    """Laser name"""

    Laser_A = "Laser A"
    Laser_B = "Laser B"
    Laser_C = "Laser C"
    Laser_D = "Laser D"
    Laser_E = "Laser E"


class Laser(AindModel):
    """Description of a laser"""

    name: LaserName = Field(..., title="Name")
    wavelength: int = Field(..., title="Wavelength (nm)")
    excitation_power: Optional[float] = Field(
        None, title="Excitation power (mW)"
    )


class Patch(AindModel):
    """Description of a patch"""

    name: PatchCordName = Field(..., title="Name")
    output_power: float = Field(..., title="Output power (uW)")


class OphysSession(AindCoreModel):
    """Description of an ophys session"""

    schema_version: str = Field(
        "0.0.1",
        description="schema version",
        title="Schema Version",
        const=True,
    )
    experimenter_full_name: str = Field(
        ...,
        description="First and last name of the experimenter.",
        title="Experimenter full name",
    )
    session_start_time: datetime = Field(..., title="Session start time")
    session_end_time: datetime = Field(..., title="Session end time")
    subject_id: int = Field(..., title="Subject ID")
    session_type: str = Field(..., title="Session type")
    stimulus_protocol_id: Optional[str] = Field(
        None, title="Stimulus protocol ID"
    )
    iacuc_protocol: Optional[str] = Field(None, title="IACUC protocol")
    rig_id: str = Field(..., title="Rig ID")
    patch_cords: List[Patch] = Field(
        ..., title="Patch cords", unique_items=True
    )
    lasers: List[Laser] = Field(..., title="Lasers", unique_items=True)
    coupling_array: List[Coupling] = Field(
        ..., title="Coupling array", unique_items=True
    )
    detectors: Optional[List[Detector]] = Field(
        None, title="Detectors", unique_items=True
    )
    notes: Optional[str] = None



""" test Ophys """

import datetime
import unittest

from pydantic import ValidationError

from aind_data_schema import OphysSession, OphysRig


class OphysTests(unittest.TestCase):
    """test ophys schemas"""

    def test_constructors(self):
        """testing constructors"""
        with self.assertRaises(ValidationError):
            r = OphysRig()

        """testing constructors"""
        with self.assertRaises(ValidationError):
            s = OphysSession()

        r = OphysRig(rig_id="12345", patch_cords=[], lasers=[])

        assert r is not None
        now = datetime.datetime.now()

        s = OphysSession(
            experimenter_full_name="alice",
            session_start_time=now,
            session_end_time=now,
            subject_id="123456",
            session_type="asdf",
            rig_id="asdf",
            lasers=[],
            coupling_array=[],
            patch_cords=[],
        )

        assert s is not None


if __name__ == "__main__":
    unittest.main()



""" test Imaging """

import datetime
import unittest

from pydantic import ValidationError

from aind_data_schema import Acquisition, Instrument
from aind_data_schema.imaging.acquisition import Axis


class ImagingTests(unittest.TestCase):
    """test imaging schemas"""

    def test_constructors(self):
        """testing constructors"""
        with self.assertRaises(ValidationError):
            a = Acquisition()

        a = Acquisition(
            experimenter_full_name="alice",
            session_start_time=datetime.datetime.now(),
            subject_id="1234",
            instrument_id="1234",
            session_end_time=datetime.datetime.now(),
            positions=[],
            channels=[],
        )

        assert a is not None

        with self.assertRaises(ValidationError):
            i = Instrument()

        i = Instrument(
            type="smartSPIM",
            location="440",
            manufacturer="LifeCanvas",
            objectives=[],
            detectors=[],
            light_sources=[],
        )

        assert i is not None

    def test_axis(self):
        """test the axis class"""
        voxel_sizes = [1, 2, 3]
        volume_sizes = [10, 10, 10]

        # test that a few work
        test_codes = ["RAS", "LSP", "RAI", "PAR"]
        for test_code in test_codes:
            axes = Axis.from_direction_code(
                test_code, voxel_sizes, volume_sizes
            )
            assert len(axes) == 3


if __name__ == "__main__":
    unittest.main()



""" schema for processing """

from __future__ import annotations

from datetime import datetime
from enum import Enum
from typing import Any, Dict, List, Optional

from pydantic import Field
from .base import AindCoreModel, AindModel


class ProcessName(Enum):
    """Data processing type labels"""

    EPHYS_PREPROCESSING = "Ephys preprocessing"
    EPHYS_POSTPROCESSING = "Ephys postprocessing"
    IMAGE_IMPORTING = "Image importing"
    IMAGE_ATLAS_ALIGNMENT = "Image atlas alignment"
    IMAGE_DESTRIPING = "Image destriping"
    IMAGE_THRESHOLDING = "Image thresholding"
    IMAGE_TILE_FUSING = "Image tile fusing"
    IMAGE_TILE_PROJECTION = "Image tile projection"
    IMAGE_TILE_STITCHING = "Image tile stitching"
    FILE_CONVERSION = "File format conversion"
    OTHER = "Other"
    SPIKE_SORTING = "Spike sorting"


class DataProcess(AindModel):
    """Description of a single processing step"""

    name: ProcessName = Field(..., title="Name")
    version: str = Field(
        ..., description="Version of the software used", title="Version"
    )
    start_date_time: datetime = Field(..., title="Start date time")
    end_date_time: datetime = Field(..., title="End date time")
    input_location: str = Field(
        ..., description="Path to data inputs", title="Input location"
    )
    output_location: str = Field(
        ..., description="Path to data outputs", title="Output location"
    )
    code_url: str = Field(
        ..., description="Path to code respository", title="Code URL"
    )
    parameters: Dict[str, Any]
    notes: Optional[str] = None


class Processing(AindCoreModel):
    """Desription of all processes run on data"""

    schema_version: str = Field(
        "0.1.0",
        description="Schema version",
        title="Schema version",
        const=True,
    )
    pipeline_version: Optional[str] = Field(
        None, description="Version of the pipeline", title="Pipeline version"
    )
    pipeline_url: Optional[str] = Field(
        None, description="URL to the pipeline code", title="Pipeline URL"
    )
    data_processes: List[DataProcess] = Field(
        ..., title="Data processing", unique_items=True
    )



""" tests for Subject """

import datetime
import unittest

import pydantic

from aind_data_schema import LightCycle, Subject
from aind_data_schema.subject import MgiAlleleId


class SubjectTests(unittest.TestCase):
    """tests for subject"""

    def test_constructors(self):
        """try building Subjects"""

        with self.assertRaises(pydantic.ValidationError):
            s = Subject()

        now = datetime.datetime.now()

        lc = LightCycle(lights_on_time=now.time(), lights_off_time=now.time())

        s = Subject(
            species="Mus musculus",
            subject_id="1234",
            sex="Male",
            date_of_birth=now.date(),
            genotype="wt",
            light_cycle=lc,
            mgi_allele_ids=[MgiAlleleId(mgi_id="12345", allele_name="adsf")],
        )

        assert s is not None


if __name__ == "__main__":
    unittest.main()



""" schema describing imaging acquisition """

from __future__ import annotations

from datetime import datetime
from enum import Enum
from typing import List, Optional

from pydantic import Field
from ..base import AindCoreModel, AindModel


class AxisName(Enum):
    """Image axis name"""

    X = "X"
    Y = "Y"
    Z = "Z"


class Direction(Enum):
    """Anatomical direction name"""

    LR = "Left_to_right"
    RL = "Right_to_left"
    AP = "Anterior_to_posterior"
    PA = "Posterior_to_anterior"
    IS = "Inferior_to_superior"
    SI = "Superior_to_inferior"
    OTHER = "Other"


class Axis(AindModel):
    """Description of an image axis"""

    name: AxisName = Field(..., title="Name")
    dimension: int = Field(
        ...,
        description="Reference axis number for stitching",
        title="Dimension",
    )
    direction: Direction = Field(
        ...,
        description="Tissue direction as the value of axis increases. If Other describe in notes.",
    )
    voxel_size: float = Field(..., title="Voxel size (um)")
    volume_size: float = Field(
        ...,
        description="Size of the volume for this dimension.",
        title="Volume size (um)",
    )

    @staticmethod
    def from_direction_code(code, voxel_sizes, volume_sizes) -> List[Axis]:
        """Convert direction codes like 'RAS' or 'PLA' into a set of axis objects"""
        direction_lookup = {
            "L": Direction.LR,
            "R": Direction.RL,
            "A": Direction.AP,
            "P": Direction.PA,
            "I": Direction.IS,
            "S": Direction.SI,
        }

        name_lookup = [AxisName.X, AxisName.Y, AxisName.Z]

        axes = []
        for i, c in enumerate(code):
            axis = Axis(
                name=name_lookup[i],
                direction=direction_lookup[c],
                dimension=i,
                voxel_size=voxel_sizes[i],
                volume_size=volume_sizes[i],
            )
            axes.append(axis)

        return axes


class Channel(AindModel):
    """Description of a channel"""

    channel: int = Field(..., title="Channel")
    enabled: Optional[bool] = Field(None, title="Enabled")
    laser_wavelength: int = Field(
        ..., title="Wavelength (nm)", units="nm", ge=300, le=1000
    )
    laser_power: float = Field(..., title="Power (mW)", units="mW", le=2000)
    filter_wheel_index: int = Field(..., title="Filter wheel index")


class Position(AindModel):
    """Description of stage position"""

    x_start_um: float
    x_end_um: float
    x_step_um: float
    y_start_um: float
    y_end_um: float
    y_step_um: float
    z_start_um: float
    z_end_um: float
    z_step_um: float


class Acquisition(AindCoreModel):
    """Description of an imaging acquisition session"""

    version: str = Field(
        "0.3.0", description="schema version", title="Version", const="True"
    )
    experimenter_full_name: str = Field(
        ...,
        description="First and last name of the experimenter.",
        title="Experimenter full name",
    )
    session_start_time: datetime = Field(..., title="Session start time")
    subject_id: int = Field(..., title="Subject ID")
    instrument_id: str = Field(..., title="Instrument ID")
    session_end_time: datetime = Field(..., title="Session end time")
    local_storage_directory: Optional[str] = Field(
        None, title="Local storage directory"
    )
    external_storage_directory: Optional[str] = Field(
        None, title="External storage directory"
    )
    tile_prefix: Optional[str] = Field(
        None,
        description="Zstacks will be named: <tile_prefix>_<x>_<y>_<wavelength>.tiff",
        title="Tile prefix",
    )
    tile_overlap_x: Optional[float] = Field(
        None, title="Tile overlap x (percent)", ge=0, le=100
    )
    tile_overlap_y: Optional[float] = Field(
        None, title="Tile overlap y (percent)", ge=0, le=100
    )
    step_size_z: Optional[float] = Field(None, title="Step size z (um)")
    axes: Optional[List[Axis]] = Field(None, title="Axes", unique_items=True)
    additional_parameters: Optional[str] = Field(
        None, title="Additional parameters"
    )
    positions: List[Position] = Field(
        ..., title="Positions", unique_items=True
    )
    channels: List[Channel] = Field(..., title="Channels", unique_items=True)
    daqs: Optional[List[dict]] = Field(None, title="DAQ", unique_items=True)
    notes: Optional[str] = Field(None, title="Notes")



""" schema for mostly mouse metadata """

from __future__ import annotations

from datetime import date, time
from enum import Enum
from typing import Optional, List

from pydantic import Field

from .base import AindCoreModel, AindModel


class Species(Enum):
    """Species latin name"""

    MUS_MUSCULUS = "Mus musculus"
    CALLITHRIX_JACCHUS = "Callithrix jacchus"
    MACACA_MULATTA = "Macaca mulatta"
    HOMO_SAPIENS = "Homo sapiens"


class Sex(Enum):
    """Subject sex name"""

    FEMALE = "Female"
    MALE = "Male"


class BackgroundStrain(Enum):
    """Animal background strain name"""

    C57BL_6J = "C57BL/6J"
    BALB_c = "BALB/c"


class LightCycle(AindModel):
    """Description of vivarium light cycle times"""

    lights_on_time: time = Field(
        ...,
        description="Time in UTC that lights were turned on",
        title="Lights on time",
    )
    lights_off_time: time = Field(
        ...,
        description="Time in UTC that lights were turned off",
        title="Lights off time",
    )


class HomeCageEnrichment(Enum):
    """Materials provided in animal home cage"""

    NONE = "none"
    RUNNING_WHEEL = "running wheel"
    SOCIAL_HOUSING = "social housing"
    PLASTIC_TUBE = "plastic tube"
    PLASTIC_SHELTER = "plastic shelter"
    OTHER = "other"


class WellnessReport(AindModel):
    """Wellness report on animal health"""

    date: date = Field(..., title="Date")
    report: str = Field(..., title="Report")


class MgiAlleleId(AindModel):
    """Mouse Genome Informatics IDs for genotype alleles"""

    allele_name: str = Field(..., title="Name")
    mgi_id: str = Field(..., title="MGI ID")


class Subject(AindCoreModel):
    """Description of a subject of data collection"""

    schema_version: str = Field(
        "0.2.2", description="schema version", title="Version", const=True
    )
    species: Species = Field(..., title="Species")
    subject_id: str = Field(
        ...,
        description="Unique identifier for the subject. If this is not a Allen LAS ID, indicate this in the Notes.",
        title="Subject ID",
    )
    sex: Sex = Field(..., title="Sex")
    date_of_birth: date = Field(..., title="Date of birth")
    genotype: str = Field(
        ...,
        description="Genotype of the animal providing both alleles",
        title="Genotype",
    )
    mgi_allele_ids: Optional[List[MgiAlleleId]] = Field(
        None, title="MGI allele ids"
    )
    background_strain: Optional[BackgroundStrain] = Field(
        None, title="Background strain"
    )
    source: Optional[str] = Field(
        None,
        description="If the subject was not bred in house, where was it acquired from.",
        title="Source",
    )
    rrid: Optional[str] = Field(
        None,
        description="RRID of mouse if acquired from supplier",
        title="RRID",
    )
    restrictions: Optional[str] = Field(
        None,
        description="Any restrictions on use or publishing based on subject source",
        title="Restrictions",
    )
    breeding_group: Optional[str] = Field(None, title="Breeding Group")
    maternal_id: Optional[str] = Field(None, title="Maternal specimen ID")
    maternal_genotype: Optional[str] = Field(None, title="Maternal genotype")
    paternal_id: Optional[str] = Field(None, title="Paternal specimen ID")
    paternal_genotype: Optional[str] = Field(None, title="Paternal genotype")
    light_cycle: Optional[LightCycle] = Field(None, title="Light cycle")
    home_cage_enrichment: Optional[HomeCageEnrichment] = Field(
        None, title="Home cage enrichment"
    )
    wellness_reports: Optional[List[WellnessReport]] = Field(
        None, title="Wellness Report"
    )
    notes: Optional[str] = Field(None, title="Notes")



""" test Procedures """

import unittest
import datetime
from pydantic import ValidationError

from aind_data_schema import Procedures
from aind_data_schema.procedures import (
    FiberImplant,
    OphysProbe,
    RetroOrbitalInjection,
    NanojectInjection,
    InjectionMaterial,
)


class ProceduresTests(unittest.TestCase):
    """test Procedures"""

    def test_constructors(self):
        """test that we can construct things"""

        with self.assertRaises(ValidationError):
            p = Procedures()

        p = Procedures(subject_id="1234")

        assert p is not None

        now = datetime.datetime.now()

        p = Procedures(
            subject_id="12345",
            injections=[
                RetroOrbitalInjection(
                    start_date=now,
                    end_date=now,
                    experimenter_full_name="tom",
                    protocol_id="134",
                    injection_materials=[
                        InjectionMaterial(
                            name="abc",
                            titer="5.46E13",
                            prep_lot_number="CT323",
                        )
                    ],
                    injection_volume=1,
                    injection_eye="Left",
                    injection_duration=1,
                    recovery_time=10,
                ),
                NanojectInjection(
                    start_date=now,
                    end_date=now,
                    experimenter_full_name="betsy",
                    protocol_id="bca",
                    injection_materials=[
                        InjectionMaterial(
                            name="abc",
                            titer="5.46E13",
                            prep_lot_number="CT323",
                        )
                    ],
                    injection_duration=1,
                    injection_coordinate_ml=1,
                    injection_coordinate_ap=1,
                    injection_coordinate_depth=1,
                    injection_angle=1,
                    injection_volume=1,
                    recovery_time=10,
                ),
            ],
            fiber_implants=[
                FiberImplant(
                    start_date=now,
                    end_date=now,
                    experimenter_full_name="Betsy",
                    protocol_id="dx.doi.org/120.123/fkjd",
                    probes=[
                        OphysProbe(
                            name="Probe A",
                            manufacturer="Tom",
                            part_number="8",
                            core_diameter=2,
                            numerical_aperture=1,
                            targeted_structure="MOp",
                            stereotactic_coordinate_ap=1,
                            stereotactic_coordinate_dv=2,
                            stereotactic_coordinate_ml=3,
                            angle=10,
                        )
                    ],
                )
            ],
        )


if __name__ == "__main__":
    unittest.main()



""" example unit test file """

import datetime
import unittest

import pydantic

from aind_data_schema import EphysRig, EphysSession
from aind_data_schema.ephys.ephys_session import (
    EphysProbe,
    Stream,
    CcfCoords,
    Coordinates3d,
)
from aind_data_schema.ephys.ephys_rig import (
    ManipulatorAngle,
    Manipulator,
    LaserModule,
)


class ExampleTest(unittest.TestCase):
    """an example test"""

    def test_constructors(self):
        """always returns true"""

        with self.assertRaises(pydantic.ValidationError):
            er = EphysRig()

        with self.assertRaises(pydantic.ValidationError):
            es = EphysSession()

        er = EphysRig(
            rig_id="1234",
            lasers=[
                LaserModule(
                    manufacturer="Hamamatsu",
                    serial_number="1234",
                    name="Laser A",
                    laser_manipulator=Manipulator(
                        manufacturer="Other",
                        serial_number="1234",
                        manipulator_angles=[
                            ManipulatorAngle(name="XY", value=1),
                            ManipulatorAngle(name="YZ", value=1),
                            ManipulatorAngle(name="XZ", value=1),
                        ],
                    ),
                )
            ],
        )

        assert er is not None

        es = EphysSession(
            experimenter_full_name="alice",
            session_start_time=datetime.datetime.now(),
            session_end_time=datetime.datetime.now(),
            subject_id="1234",
            session_type="Test",
            rig_id="1234",
            probe_streams=[
                Stream(
                    stream_start_time=datetime.datetime.now(),
                    stream_stop_time=datetime.datetime.now(),
                    probes=[
                        EphysProbe(
                            name="Probe A",
                            tip_targeted_structure="VISl4",
                            targeted_ccf_coordinates=CcfCoords(
                                ml="1", ap="1", dv="1"
                            ),
                            manipulator_coordinates=Coordinates3d(
                                x="1", y="1", z="1"
                            ),
                        )
                    ],
                    lasers=[],
                )
            ],
        )

        assert es is not None


if __name__ == "__main__":
    unittest.main()



""" generic base class with supporting validators and fields for basic AIND schema """

from pydantic import BaseModel, Extra
from pydantic.fields import ModelField

import urllib.parse
import os
import inspect

DESCRIBED_BY_BASE_URL = "https://raw.githubusercontent.com/AllenNeuralDynamics/aind-data-schema/main/src/"


def build_described_by(cls, base_url=DESCRIBED_BY_BASE_URL):
    """construct a pydantic Field that refers to a specific file"""

    # get the filename of the class
    filename = inspect.getfile(cls)

    # strip off local directories
    package_index = filename.rfind("aind_data_schema")
    filename = filename[package_index:]

    # forward slashes
    filename = filename.replace(os.sep, "/")

    described_by = urllib.parse.urljoin(base_url, filename)

    return described_by


class AindModel(BaseModel, extra=Extra.forbid):
    """BaseModel that disallows extra fields"""

    pass


class AindCoreModel(AindModel):
    """Generic base class to hold common fields/validators/etc for all basic AIND schema"""

    describedBy: str

    def __init_subclass__(cls, optional_fields=None, **kwargs):
        """add the describedby field to all subclasses"""
        super().__init_subclass__(**kwargs)

        value = build_described_by(cls)
        field = ModelField.infer(
            name="describedBy",
            value=value,
            annotation=str,
            class_validators=None,
            config=cls.__config__,
        )
        field.field_info.const = True
        cls.__fields__.update({"describedBy": field})



""" example procedures """

import datetime
from aind_data_schema import Procedures
from aind_data_schema.procedures import (
    Craniotomy,
    TrainingProtocol,
    FiberImplant,
    OphysProbe,
)

t = datetime.datetime(2022, 11, 22, 8, 43, 00)

p = Procedures(
    subject_id="12345",
    training_protocols=[
        TrainingProtocol(
            protocol_id="dx.doi.org/10.12345/protocols.io.12345",
            training_protocol_start_date=t.date(),
            training_protocol_end_date=t.date(),
        )
    ],
    craniotomies=[
        Craniotomy(
            experimenter_full_name="jane doe",
            protocol_id="dx.doi.org/10.12345/protocols.io.12345",
            craniotomy_coordinates_ap=0,
            craniotomy_coordinates_ml=1,
            craniotomy_size=2,
            protective_material="Duragel",
            animal_weight=10,
            start_date=t.date(),
            end_date=t.date(),
        )
    ],
    fiber_implants=[
        FiberImplant(
            experimenter_full_name="john doe",
            protocol_id="dx.doi.org/10.12345/protocols.io.12345",
            animal_weight=34,
            start_date=t.date(),
            end_date=t.date(),
            probes=[
                OphysProbe(
                    name="Probe A",
                    manufacturer="acmecorp",
                    part_number="12345",
                    core_diameter=10,
                    numerical_aperture=1,
                    targeted_structure="VISp",
                    stereotactic_coordinate_ap=0,
                    stereotactic_coordinate_ml=0,
                    stereotactic_coordinate_dv=0,
                    angle=0,
                )
            ],
        )
    ],
)

with open("procedures.json", "w") as f:
    f.write(p.json(indent=3))



""" example ExaSPIM instrument """

from aind_data_schema.imaging import instrument
from aind_data_schema.device import DaqChannel


inst = instrument.Instrument(
    instrument_id="exaSPIM1-1",
    type="exaSPIM",
    manufacturer="Custom",
    location="440 Westlake",
    objectives=[
        instrument.Objective(
            numerical_aperture=0.305,
            magnification=5.0,
            immersion="air",
            manufacturer="Other",
            model="JM_DIAMOND 5.0X/1.3",
            serial_number="Unknown",
            notes="manufacturer collaboration between Schneider-Kreuznach and Vieworks",
        ),
    ],
    detectors=[
        instrument.Detector(
            type="Camera",
            data_interface="Coax",
            cooling="air",
            manufacturer="Vieworks",
            model="VNP-604MX",
            serial_number="MB151BAY001",
        ),
    ],
    light_sources=[
        instrument.Lightsource(
            type="laser",
            coupling="SMF",
            wavelength=405,
            max_power=200,
            serial_number="LAS-08307",
            manufacturer="Oxxius",
            notes="Housed in commercial laser combiner",
            daq_channel=DaqChannel(index=6, type="Analog Output"),
        ),
        instrument.Lightsource(
            type="laser",
            coupling="SMF",
            wavelength=488,
            max_power=200,
            serial_number="LAS-08308",
            manufacturer="Oxxius",
            notes="Housed in commercial laser combiner",
            daq_channel=DaqChannel(index=3, type="Analog Output"),
        ),
        instrument.Lightsource(
            type="laser",
            coupling="SMF",
            wavelength=561,
            max_power=200,
            serial_number="539251",
            manufacturer="Oxxius",
            notes="Housed in commercial laser combiner",
            daq_channel=DaqChannel(index=5, type="Analog Output"),
        ),
        instrument.Lightsource(
            type="laser",
            coupling="SMF",
            wavelength=638,
            max_power=200,
            serial_number="LAS-08309",
            manufacturer="Oxxius",
            notes="Housed in commercial laser combiner",
            daq_channel=DaqChannel(index=4, type="Analog Output"),
        ),
    ],
    fluorescence_filters=[
        instrument.Filter(
            type="Multiband",
            manufacturer="Chroma",
            diameter=44.05,
            thickness=1.0,
            model="ZET405/488/561/640mv2",
            notes="Custom made filter",
            filter_wheel_index=0,
            serial_number="Unknown-0",
        )
    ],
    daqs=[
        instrument.DAQ(
            model="PCIe-6738",
            manufacturer="National Instruments",
            device_name="Dev2",
            update_frequency="10000",
            number_active_channels=7,
            serial_number="Unknown",
        )
    ],
    scanning_stages=[
        instrument.ScanningStage(
            stage_axis_direction="Detection axis",
            stage_axis_name="X",
            travel=1000,
            model="MS-8000",
            manufacturer="Applied Scientific Instrumentation",
            daq_channel=DaqChannel(index=2, type="Analog Output"),
            serial_number="Unknown",
        ),
        instrument.ScanningStage(
            stage_axis_direction="Perpendicular axis",
            stage_axis_name="Y",
            travel=1000,
            model="MS-8000",
            manufacturer="Applied Scientific Instrumentation",
            serial_number="Unknown",
        ),
        instrument.ScanningStage(
            stage_axis_direction="Illumination axis",
            stage_axis_name="Z",
            travel=100,
            model="LS-100",
            manufacturer="Applied Scientific Instrumentation",
            serial_number="Unknown",
        ),
    ],
    additional_devices=[
        instrument.AdditionalImagingDevice(
            type="Tunable lens",
            manufacturer="Optotune",
            model="EL-16-40-TC-VIS-20D-C",
            serial_number="01",
            daq_channel=DaqChannel(index=0, type="Analog Output"),
        ),
        instrument.AdditionalImagingDevice(
            type="Rotation mount",
            manufacturer="Thorlabs",
            model="K10CR1",
            serial_number="01",
        ),
        instrument.AdditionalImagingDevice(
            type="Laser combiner",
            manufacturer="Oxxius",
            model="L6Cc",
            serial_number="L6CC-00513",
        ),
    ],
    optical_tables=[
        instrument.OpticalTable(
            length=36,
            width=48,
            vibration_control=True,
            model="VIS3648-PG2-325A",
            manufacturer="MKS Newport",
            serial_number="Unknown",
        )
    ],
    com_ports=[
        instrument.Com(
            hardware_name="Laser Launch",
            com_port="COM2",
        ),
        instrument.Com(
            hardware_name="ASI Tiger",
            com_port="COM5",
        ),
    ],
    humidity_control=False,
    temperature_control=False,
)

with open("exaspim_instrument.json", "w") as f:
    f.write(inst.json(indent=3))



""" example SmartSPIM instrument """

from aind_data_schema.imaging.instrument import (
    Instrument,
    Objective,
    Detector,
    Lightsource,
    MotorizedStage,
    ScanningStage,
    Com,
    Filter,
    OpticalTable,
    AdditionalImagingDevice,
)

inst = Instrument(
    instrument_id="SmartSPIM1-1",
    type="smartSPIM",
    manufacturer="LifeCanvas",
    location="440 Westlake",
    objectives=[
        Objective(
            numerical_aperture=0.2,
            magnification=3.6,
            immersion="multi",
            manufacturer="Thorlabs",
            model="TL4X-SAP",
            serial_number="Unknown",
            notes="Thorlabs TL4X-SAP with LifeCanvas dipping cap and correction optics.",
        ),
    ],
    detectors=[
        Detector(
            type="Camera",
            data_interface="USB",
            cooling="air",
            manufacturer="Hamamatsu",
            model="C14440-20UP",
            serial_number="001107",
        ),
    ],
    light_sources=[
        Lightsource(
            type="laser",
            coupling="SMF",
            wavelength=488,
            max_power=150,
            serial_number="VL01222A11",
            manufacturer="Vortran",
            model="Stradus",
            notes="All lasers controlled via Vortran VersaLase System",
        ),
        Lightsource(
            type="laser",
            coupling="SMF",
            wavelength=561,
            max_power=150,
            serial_number="417927",
            manufacturer="Coherent Scientific",
            model="Obis",
            notes="All lasers controlled via Vortran VersaLase System",
        ),
        Lightsource(
            type="laser",
            coupling="SMF",
            wavelength=647,
            max_power=160,
            serial_number="VL01222A10",
            manufacturer="Vortran",
            model="Stradus",
            notes="All lasers controlled via Vortran VersaLase System",
        ),
    ],
    motorized_stages=[
        MotorizedStage(
            model="LS-100",
            manufacturer="Applied Scientific Instrumentation",
            serial_number="Unknown-1",
            travel=100,
            notes="Focus stage",
        ),
        MotorizedStage(
            model="L12-20F-4",
            manufacturer="IR Robot Co",
            serial_number="Unknown-5",
            travel=41,
            notes="Cylindrical lens #1",
        ),
        MotorizedStage(
            model="L12-20F-4",
            manufacturer="IR Robot Co",
            serial_number="Unknown-6",
            travel=41,
            notes="Cylindrical lens #2",
        ),
        MotorizedStage(
            model="L12-20F-4",
            manufacturer="IR Robot Co",
            serial_number="Unknown-7",
            travel=41,
            notes="Cylindrical lens #3",
        ),
        MotorizedStage(
            model="L12-20F-4",
            manufacturer="IR Robot Co",
            serial_number="Unknown-8",
            travel=41,
            notes="Cylindrical lens #4",
        ),
    ],
    scanning_stages=[
        ScanningStage(
            model="LS-50",
            manufacturer="Applied Scientific Instrumentation",
            serial_number="Unknown-2",
            stage_axis_direction="Detection axis",
            stage_axis_name="Z",
            travel=50,
            notes="Sample stage Z",
        ),
        ScanningStage(
            model="LS-50",
            manufacturer="Applied Scientific Instrumentation",
            serial_number="Unknown-3",
            stage_axis_direction="Illumination axis",
            stage_axis_name="X",
            travel=50,
            notes="Sample stage X",
        ),
        ScanningStage(
            model="LS-50",
            manufacturer="Applied Scientific Instrumentation",
            serial_number="Unknown-4",
            stage_axis_direction="Perpendicular axis",
            stage_axis_name="Y",
            travel=50,
            notes="Sample stage Y",
        ),
    ],
    optical_tables=[
        OpticalTable(
            model="VIS2424-IG2-125A",
            length=24,
            width=24,
            vibration_control=False,
            manufacturer="MKS Newport",
            serial_number="Unknown",
        )
    ],
    humidity_control=False,
    temperature_control=False,
    com_ports=[
        Com(
            hardware_name="Laser Launch",
            com_port="COM3",
        ),
        Com(
            hardware_name="ASI Tiger",
            com_port="COM5",
        ),
        Com(hardware_name="MightyZap", com_port="COM4"),
    ],
    fluorescence_filters=[
        Filter(
            type="Band pass",
            manufacturer="Semrock",
            diameter=25,
            thickness=2.0,
            model="FF03-525/50-25",
            filter_wheel_index=0,
            serial_number="Unknown-1",
        ),
        Filter(
            type="Band pass",
            manufacturer="Semrock",
            diameter=25,
            thickness=2.0,
            model="FF01-600/52-25",
            filter_wheel_index=1,
            serial_number="Unknown-2",
        ),
        Filter(
            type="Band pass",
            manufacturer="Chroma",
            diameter=25,
            thickness=2.0,
            model="ET690/50m",
            filter_wheel_index=2,
            serial_number="Unknown-3",
        ),
    ],
    additional_devices=[
        AdditionalImagingDevice(
            type="Other",
            manufacturer="Optotune",
            model="EL-16-40-TC",
            serial_number="Unknown-1",
        ),
        AdditionalImagingDevice(
            type="Other",
            manufacturer="Optotune",
            model="EL-16-40-TC",
            serial_number="Unknown-2",
        ),
    ],
)

with open("aibs_smartspim_instrument.json", "w") as f:
    f.write(inst.json(indent=3))



""" ephys rig schemas """

from __future__ import annotations

from datetime import datetime
from enum import Enum
from typing import List, Optional

from pydantic import Field

from ..device import Device, DAQ
from ..base import AindCoreModel, AindModel


class HarpDeviceName(Enum):
    """Harp device name"""

    BEHAVIOR = "Behavior"
    CAMERA_CONTROLLER = "Camera Controller"
    LOAD_CELLS = "Load Cells"
    SOUND_BOARD = "Sound Board"
    TIMESTAMP_GENERATOR = "Timestamp Generator"
    INPUT_EXPANDER = "Input Expander"


class HarpDevice(Device):
    """Describes a Harp device"""

    name: HarpDeviceName = Field(..., title="Name")
    device_version: str = Field(..., title="Device version")


class CameraName(Enum):
    """Camera name"""

    BODY_CAMERA = "Body Camera"
    EYE_CAMERA = "Eye Camera"
    FACE_CAMERA = "Face Camera"


class Camera(Device):
    """Description of camera"""

    name: CameraName = Field(..., title="Name")
    position_x: float = Field(..., title="Position X")
    position_y: float = Field(..., title="Position Y")
    position_z: float = Field(..., title="Position Z")
    angle_pitch: float = Field(
        ..., title="Angle pitch (deg)", units="deg", ge=0, le=360
    )
    angle_yaw: float = Field(
        ..., title="Angle yaw (deg)", units="deg", ge=0, le=360
    )
    angle_roll: float = Field(
        ..., title="Angle roll (deg)", units="deg", ge=0, le=360
    )
    recording_software: Optional[str] = Field(None, title="Recording software")
    recording_software_version: Optional[str] = Field(
        None, title="Recording software version"
    )


class MousePlatform(Device):
    """Description of a mouse platform"""

    surface_material: Optional[str] = Field(None, title="Surface material")


class Disc(MousePlatform):
    """Description of a running disc"""

    platform_type: str = Field("Disc", title="Platform type", const=True)
    radius: float = Field(..., title="Radius (cm)", units="cm", ge=0)
    date_surface_replaced: Optional[datetime] = Field(
        None, title="Date surface replaced"
    )


class Tube(MousePlatform):
    """Description of a tube platform"""

    platform_type: str = Field("Tube", title="Platform type", const=True)
    diameter: float = Field(..., title="Diameter (cm)", units="cm", ge=0)


class Treadmill(MousePlatform):
    """Descrsiption of treadmill platform"""

    platform_type: str = Field("Treadmill", title="Platform type", const=True)


class AngleName(Enum):
    """Euler angle name"""

    XY = "XY"
    XZ = "XZ"
    YZ = "YZ"


class ManipulatorAngle(AindModel):
    """Description of manipulator angle"""

    name: AngleName = Field(..., title="AngleName")
    value: float = Field(..., title="Value (deg)", units="deg")


class Manipulator(Device):
    """Description of manipulator"""

    manipulator_angles: List[ManipulatorAngle] = Field(
        ..., title="Manipulator angles", unique_items=True
    )


class LaserName(Enum):
    """Laser name"""

    LASER_A = "Laser A"
    LASER_B = "Laser B"
    LASER_C = "Laser C"
    LASER_D = "Laser D"


class LaserModule(Device):
    """Description of lasers used in ephys recordings"""

    name: LaserName = Field(..., title="Laser Name")
    wavelength: Optional[int] = Field(
        None, title="Wavelength (nm)", units="nm", ge=300, le=1000
    )
    maximum_power: Optional[float] = Field(
        None, title="Maximum power (mW)", units="mW"
    )
    coupling_efficiency: Optional[float] = Field(
        None,
        title="Coupling efficiency (percent)",
        units="percent",
        ge=0,
        le=100,
    )
    calibration_data: Optional[str] = Field(
        None, description="path to calibration data", title="Calibration data"
    )
    calibration_date: Optional[datetime] = Field(
        None, title="Calibration date"
    )
    laser_manipulator: Manipulator = Field(..., title="Manipulator")


class Monitor(Device):
    """Description of a visual monitor"""

    refresh_rate: int = Field(
        ..., title="Refresh rate (Hz)", units="Hz", ge=60
    )
    width: int = Field(..., title="Width (pixels)", units="pixels")
    height: int = Field(..., title="Height (pixels)", units="pixels")
    viewing_distance: float = Field(
        ..., title="Viewing distance (cm)", units="cm"
    )
    position_x: float = Field(..., title="Position X")
    position_y: float = Field(..., title="Position Y")
    position_z: float = Field(..., title="Position Z")
    angle_pitch: float = Field(
        ..., title="Angle pitch (deg)", units="deg", ge=0, le=360
    )
    angle_yaw: float = Field(
        ..., title="Angle yaw (deg)", units="deg", ge=0, le=360
    )
    angle_roll: float = Field(
        ..., title="Angle roll (deg)", units="deg", ge=0, le=360
    )
    contrast: int = Field(
        ...,
        description="Monitor's contrast setting",
        title="Contrast (percent)",
        units="percent",
        ge=0,
        le=100,
    )
    brightness: int = Field(
        ...,
        description="Monitor's brightness setting",
        title="Brightness",
        ge=0,
        le=100,
    )


class ProbeName(Enum):
    """Probe name"""

    PROBE_A = "Probe A"
    PROBE_B = "Probe B"
    PROBE_C = "Probe C"
    PROBE_D = "Probe D"
    PROBE_E = "Probe E"
    PROBE_F = "Probe F"
    PROBE_G = "Probe G"
    PROBE_H = "Probe H"
    PROBE_I = "Probe I"
    PROBE_J = "Probe J"


class ProbeModel(Enum):
    """Probe model name"""

    NP1 = "Neuropixels 1.0"
    NP_UHD_FIXED = "Neuropixels UHD (Fixed)"
    NP_UHD_SWITCHABLE = "Neuropixels UHD (Switchable)"
    NP2_SINGLE_SHANK = "Neuropixels 2.0 (Single Shank)"
    NP2_MULTI_SHANK = "Neuropixels 2.0 (Multi Shank)"
    NP2_QUAD_BASE = "Neuropixels 2.0 (Quad Base)"
    NP_OPTO_DEMONSTRATOR = "Neuropixels Opto (Demonstrator)"
    MI_ULED_PROBE = "Michigan uLED Probe (Version 1)"
    MP_PHOTONIC_V1 = "MPI Photonic Probe (Version 1)"


class EphysProbe(Device):
    """Description of an ephys probe"""

    name: ProbeName = Field(..., title="Name")
    model: ProbeModel = Field(..., title="Model")
    probe_manipulator: Manipulator = Field(..., title="Manipulator")
    calibration_data: str = Field(
        ..., title="Calibration data", description="Path to calibration data"
    )
    calibration_date: Optional[datetime] = Field(
        None, title="Calibration date"
    )


class EphysRig(AindCoreModel):
    """Description of an ephys rig"""

    describedBy: str = Field(
        "https://github.com/AllenNeuralDynamics/data_schema/blob/main/schemas/ephys/ephys_rig.py",
        description="The URL reference to the schema.",
        title="Described by",
        const=True,
    )
    schema_version: str = Field(
        "0.4.2", description="schema version", title="Version", const=True
    )
    rig_id: str = Field(
        ..., description="room_stim apparatus_version", title="Rig ID"
    )
    probes: Optional[List[EphysProbe]] = Field(
        None, title="Ephys probes", unique_items=True
    )
    cameras: Optional[List[Camera]] = Field(
        None, title="Cameras", unique_items=True
    )
    lasers: Optional[List[LaserModule]] = Field(
        None, title="Lasers", unique_items=True
    )
    visual_monitors: Optional[List[Monitor]] = Field(
        None, title="Visual monitor", unique_items=True
    )
    mouse_platform: Optional[MousePlatform] = Field(
        None, title="Mouse platform"
    )
    harp_devices: Optional[List[HarpDevice]] = Field(
        None, title="Harp devices"
    )
    daq: Optional[DAQ] = Field(None, title="DAQ")



""" testing examples """

import glob
import unittest
from pathlib import Path
import importlib
import sys
from unittest.mock import patch, mock_open
import logging

EXAMPLES_DIR = Path(__file__).parents[1] / "examples"


class ExampleTests(unittest.TestCase):
    """tests for examples"""

    def test_examples(self):
        """run through each example, compare to rendered json"""

        for example_file in glob.glob(f"{EXAMPLES_DIR}/*.py"):
            logging.debug(f"testing {example_file}")

            json_file = example_file.replace(".py", ".json")

            with open(json_file, "r") as f:
                target_data = f.read().replace("\r\n", "\n")

            spec = importlib.util.spec_from_file_location(
                "test_module", example_file
            )
            module = importlib.util.module_from_spec(spec)
            sys.modules["test_module"] = module

            with patch(
                "test_module.open", new_callable=mock_open
            ) as mocked_file:
                spec.loader.exec_module(module)
                h = mocked_file()
                h.write.assert_called_once_with(target_data)


if __name__ == "__main__":
    unittest.main()



""" tests for Subject """
import unittest
from aind_data_schema import Procedures


class BaseTests(unittest.TestCase):
    """tests for the base module"""

    def test_described_by(self):
        """test that described by works"""

        p = Procedures(subject_id="1234")

        self.assertEqual(
            p.describedBy,
            "https://raw.githubusercontent.com/AllenNeuralDynamics/aind-data-schema/main/src/aind_data_schema/procedures.py",
        )


if __name__ == "__main__":
    unittest.main()



""" imports for AindModel subclasses
"""

__version__ = "0.3.8"


from .data_description import (
    DataDescription,
    DerivedDataDescription,
    RawDataDescription,
    Funding,
)
from .ephys.ephys_rig import EphysRig
from .ephys.ephys_session import EphysSession
from .imaging.acquisition import Acquisition, Axis
from .imaging.instrument import Instrument
from .procedures import Procedures
from .processing import DataProcess, Processing
from .subject import LightCycle, Subject
from .ophys.ophys_rig import OphysRig
from .ophys.ophys_session import OphysSession

__all__ = [
    "DataDescription",
    "RawDataDescription",
    "DerivedDataDescription",
    "Funding",
    "Procedures",
    "EphysRig",
    "EphysSession",
    "Subject",
    "LightCycle",
    "DataProcess",
    "Processing",
    "Acquisition",
    "Instrument",
    "OphysRig",
    "OphysSession",
    "Axis",
]



""" schema describing imaging instrument """

from __future__ import annotations

from datetime import date
from enum import Enum
from typing import List, Optional

from pydantic import Field
from ..base import AindCoreModel, AindModel

from ..device import Device, Manufacturer, DAQ


class InstrumentType(Enum):
    """Instrument type name"""

    MESOSPIM = "mesoSPIM"
    EXASPIM = "exaSPIM"
    DISPIM = "diSPIM"
    SMARTSPIM = "smartSPIM"
    CONFOCAL = "Confocal"
    TWO_PHOTON = "Two photon"
    OTHER = "Other"


class Com(AindModel):
    """Description of a communication system"""

    hardware_name: str = Field(..., title="Controlled hardware device")
    com_port: str = Field(..., title="COM port")


class CameraType(Enum):
    """Camera type name"""

    CAMERA = "Camera"
    PMT = "PMT"
    OTHER = "other"


class DataInterface(Enum):
    """Data interface name"""

    USB = "USB"
    CAMERALINK = "CameraLink"
    COAX = "Coax"
    OTHER = "other"


class Cooling(Enum):
    """Cooling medium name"""

    AIR = "air"
    WATER = "water"


class Detector(Device):
    """Description of a detector device"""

    type: CameraType = Field(..., title="Detector type")
    data_interface: DataInterface = Field(..., title="Data interface")
    cooling: Cooling = Field(..., title="Cooling")


class FilterType(Enum):
    """Filter type name"""

    LONG_PASS = "Long pass"
    BAND_PASS = "Band pass"
    SHORT_PASS = "Short pass"
    MULTIBAND = "Multiband"


class Filter(Device):
    """Description of a filter device"""

    type: FilterType = Field(..., title="Filter Type")
    diameter: float = Field(..., title="Size (mm)", ge=0)
    thickness: float = Field(..., title="Size (mm)", ge=0)
    filter_wheel_index: int = Field(..., title="Filter wheel index")
    description: Optional[str] = Field(
        None, description="Where/how filter is being used", title="Description"
    )


class LightsourceType(Enum):
    """Light source type name"""

    LAMP = "lamp"
    LASER = "laser"
    LED = "LED"
    OTHER = "other"


class Coupling(Enum):
    """Coupling method name"""

    FREE_SPACE = "Free-space"
    SMF = "SMF"
    MMF = "MMF"
    OTHER = "other"


class Lightsource(Device):
    """Description of lightsource device"""

    type: LightsourceType = Field(..., title="Lightsource Type")
    coupling: Coupling = Field(..., title="Coupling")
    wavelength: float = Field(
        ..., title="Wavelength (nm)", units="nm", ge=300, le=1000
    )
    max_power: float = Field(..., title=" Maximum power (mW)", units="mW")


class Immersion(Enum):
    """Immersion media name"""

    AIR = "air"
    WATER = "water"
    OIL = "oil"
    MULTI = "multi"
    OTHER = "other"


class Objective(Device):
    """Description of an objective device"""

    numerical_aperture: float = Field(..., title="Numerical aperture (in air)")
    magnification: float = Field(..., title="Magnification")
    immersion: Immersion = Field(..., title="Immersion")


class ImagingDeviceType(Enum):
    """Imaginge device type name"""

    DIFFUSER = "Diffuser"
    GALVO = "Galvo"
    BEAM_EXPANDER = "Beam expander"
    LASER_COUPLER = "Laser coupler"
    PRISM = "Prism"
    OBJECTIVE = "Objective"
    SLIT = "Slit"
    TUNABLE_LENS = "Tunable lens"
    ROTATION_MOUNT = "Rotation mount"
    LASER_COMBINER = "Laser combiner"
    OTHER = "Other"


class AdditionalImagingDevice(Device):
    """Description of additional devices"""

    type: ImagingDeviceType = Field(..., title="Device type")


class StageAxisDirection(Enum):
    """Direction of motion for motorized stage"""

    DETECTION_AXIS = "Detection axis"
    ILLUMINATION_AXIS = "Illumination axis"
    PERPENDICULAR_AXIS = "Perpendicular axis"


class StageAxisName(Enum):
    """Axis names for motorized stages as configured by hardware"""

    X = "X"
    Y = "Y"
    Z = "Z"


class MotorizedStage(Device):
    """Description of motorized stage"""

    travel: float = Field(..., title="Travel of device (mm)", units="mm")


class ScanningStage(MotorizedStage):
    """Description of a scanning motorized stages"""

    stage_axis_direction: StageAxisDirection = Field(
        ..., title="Direction of stage axis"
    )
    stage_axis_name: StageAxisName = Field(..., title="Name of stage axis")


class OpticalTable(Device):
    """Description of Optical Table"""

    length: Optional[float] = Field(
        None, title="Length (inches)", units="inches", ge=0
    )
    width: Optional[float] = Field(
        None, title="Width (inches)", units="inches", ge=0
    )
    vibration_control: Optional[bool] = Field(None, title="Vibration control")


class Instrument(AindCoreModel):
    """Description of an instrument, which is a collection of devices"""

    version: str = Field(
        "0.4.1", description="schema version", title="Version", const=True
    )
    instrument_id: Optional[str] = Field(
        None,
        description="unique identifier for this instrument configuration",
        title="Instrument ID",
    )
    type: InstrumentType = Field(..., title="Instrument type")
    location: str = Field(..., title="Instrument location")
    manufacturer: Manufacturer = Field(..., title="Instrument manufacturer")
    temperature_control: Optional[bool] = Field(
        None, title="Temperature control"
    )
    humidity_control: Optional[bool] = Field(None, title="Humidity control")
    optical_tables: List[OpticalTable] = Field(None, title="Optical table")
    objectives: List[Objective] = Field(
        ..., title="Objectives", unique_items=True
    )
    detectors: List[Detector] = Field(
        ..., title="Detectors", unique_items=True
    )
    light_sources: List[Lightsource] = Field(
        ..., title="Light sources", unique_items=True
    )
    fluorescence_filters: Optional[List[Filter]] = Field(
        None, title="Fluorescence filters", unique_items=True
    )
    motorized_stages: Optional[List[MotorizedStage]] = Field(
        None, title="Motorized stages", unique_items=True
    )
    scanning_stages: Optional[List[ScanningStage]] = Field(
        None, title="Scanning motorized stages", unique_items=True
    )
    daqs: Optional[List[DAQ]] = Field(None, title="DAQ", unique_items=True)
    additional_devices: Optional[List[AdditionalImagingDevice]] = Field(
        None, title="Additional devices", unique_items=True
    )
    calibration_date: Optional[date] = Field(
        None,
        description="Date of most recent calibration",
        title="Calibration date",
    )
    calibration_data: Optional[str] = Field(
        None,
        description="Path to calibration data from most recent calibration",
        title="Calibration data",
    )
    com_ports: Optional[List[Com]] = Field(
        None, title="COM ports", unique_items=True
    )
    notes: Optional[str] = None



""" example data description """

from aind_data_schema import RawDataDescription
from aind_data_schema.data_description import Funding
from datetime import date, time

d = RawDataDescription(
    modality="ecephys",
    subject_id="12345",
    creation_date=date(2022, 2, 21),
    creation_time=time(16, 30, 1),
    institution="AIND",
    funding_source=[Funding(funder="AIND")],
)

with open("data_description.json", "w") as f:
    f.write(d.json(indent=3))



""" schema for various Devices """

from pydantic import Field
from enum import Enum
from typing import Optional
from .base import AindModel


class Manufacturer(Enum):
    """Device manufacturer name"""

    THORLABS = "Thorlabs"
    OPTOTUNE = "Optotune"
    CAMBRIDGE_TECHNOLOGY = "Cambridge Technology"
    NIKON = "Nikon"
    EDMUND_OPTICS = "Edmund Optics"
    EALING = "Ealing"
    HAMAMATSU = "Hamamatsu"
    OLYMPUS = "Olympus"
    LEICA = "Leica"
    LIFECANVAS = "LifeCanvas"
    CUSTOM = "Custom"
    CHROMA = "Chroma"
    SEMROCK = "Semrock"
    NEUROPIXELS = "Neuropixels"
    MPI = "MPI"
    VORTRAN = "Vortran"
    COHERENT_SCIENTIFIC = "Coherent Scientific"
    MKS_NEWPORT = "MKS Newport"
    ASI = "Applied Scientific Instrumentation"
    MIGHTY_ZAP = "IR Robot Co"
    VIEWORKS = "Vieworks"
    OXXIUS = "Oxxius"
    NATIONAL_INSTRUMENTS = "National Instruments"
    OTHER = "Other"


class DaqChannelType(Enum):
    """DAQ Channel type"""

    DO = "Digital Output"
    AO = "Analog Output"


class DaqChannel(AindModel):
    """Description of a DAQ Channel"""

    index: int = Field(..., title="index")
    type: DaqChannelType = Field(..., title="DAQ channel type")


class DeviceBase(AindModel):
    """Description of a general device"""

    manufacturer: Manufacturer = Field(..., title="Manufacturer")
    serial_number: str = Field(..., title="Serial number")
    model: Optional[str] = Field(None, title="Model")
    notes: Optional[str] = Field(None, title="Notes")


class Device(DeviceBase):
    """Description of a device that communicates with a DAQ"""

    daq_channel: Optional[DaqChannel] = Field(None, title="DAQ channel")


class DAQ(DeviceBase):
    """Description of DAQ system"""

    device_name: str = Field(..., title="PC device name")
    update_frequency: float = Field(
        ..., title="DAQ update frequency (Hz)", units="Hz"
    )
    number_active_channels: int = Field(..., title="Number of active channels")



""" test processing """

import unittest

import pydantic

from aind_data_schema import Processing


class ProcessingTest(unittest.TestCase):
    """tests for processing schema"""

    def test_constructors(self):
        """test creation"""

        with self.assertRaises(pydantic.ValidationError):
            p = Processing()

        p = Processing(data_processes=[])

        assert p is not None


if __name__ == "__main__":
    unittest.main()



""" Generic metadata classes for data """

from __future__ import annotations

import re
from datetime import date, datetime, time
from enum import Enum
from typing import Optional, List

from pydantic import Field
from .base import AindCoreModel, AindModel


class RegexParts(Enum):
    """regular expression components to be re-used elsewhere"""

    DATE = r"\d{4}-\d{2}-\d{2}"
    TIME = r"\d{2}-\d{2}-\d{2}"


class DataRegex(Enum):
    """regular expression patterns for different kinds of data and their properties"""

    DATA = f"^(?P<label>.+?)_(?P<c_date>{RegexParts.DATE.value})_(?P<c_time>{RegexParts.TIME.value})$"
    RAW_DATA = f"^(?P<modality>.+?)_(?P<subject_id>.+?)_(?P<c_date>{RegexParts.DATE.value})_(?P<c_time>{RegexParts.TIME.value})$"
    DERIVED_DATA = f"^(?P<input>.+?_{RegexParts.DATE.value}_{RegexParts.TIME.value})_(?P<process_name>.+?)_(?P<c_date>{RegexParts.DATE.value})_(?P<c_time>{RegexParts.TIME.value})"
    NO_UNDERSCORES = "^[^_]+$"


class DataLevel(Enum):
    """Data level name"""

    RAW_DATA = "raw data"
    DERIVED_DATA = "derived data"


class Institution(Enum):
    """Institution name"""

    AIND = "AIND"
    AIBS = "AIBS"
    HUST = "HUST"


class Group(Enum):
    """Data collection group name"""

    EPHYS = "ephys"
    OPHYS = "ophys"
    MSMA = "MSMA"
    BEHAVIOR = "behavior"


class Modality(Enum):
    """Data collection modality name"""

    ECEPHYS = "ecephys"
    EXASPIM = "ExASPIM"
    SMARTSPIM = "SmartSPIM"
    MESOSPIM = "mesoSPIM"
    OPHYS = "ophys"
    FMOST = "fMOST"


def datetime_to_name_string(d, t):
    """Take a date and time object, format it a as string"""
    ds = d.strftime("%Y-%m-%d")
    ts = t.strftime("%H-%M-%S")
    return f"{ds}_{ts}"


def datetime_from_name_string(d, t):
    """Take date and time strings, generate date and time objects"""
    return (
        datetime.strptime(d, "%Y-%m-%d").date(),
        datetime.strptime(t, "%H-%M-%S").time(),
    )


def build_data_name(label, creation_date, creation_time):
    """Construct a valid data description name"""
    dt_str = datetime_to_name_string(creation_date, creation_time)
    return f"{label}_{dt_str}"


class Funding(AindModel):
    """Description of funding sources"""

    funder: str = Field(..., title="Funder")
    grant_number: Optional[str] = Field(None, title="Grant number")
    fundee: Optional[str] = Field(
        None, title="Fundee", description="Person(s) funded by this mechanism"
    )


class DataDescription(AindCoreModel):
    """Description of a logical collection of data files"""

    schema_version: str = Field("0.3.0", title="Schema Version", const=True)
    license: str = Field("CC-BY-4.0", title="License", const=True)

    creation_time: time = Field(
        ...,
        description="Time in UTC that data files were created, used to uniquely identify the data",
        title="Creation Time",
    )
    creation_date: date = Field(
        ...,
        description="Date in UTC that data files were created, used to uniquely identify the data",
        title="Creation Date",
    )
    name: str = Field(
        ...,
        description="Name of data, conventionally also the name of the directory containing all data and metadata",
        title="Name",
    )
    institution: Institution = Field(
        ...,
        description="An established society, corporation, foundation or other organization that collected this data",
        title="Institution",
    )
    funding_source: List[Funding] = Field(
        ...,
        title="Funding source",
        description="Funding sources. If internal label as Institution.",
    )
    data_level: DataLevel = Field(
        ...,
        description="level of processing that data has undergone",
        title="Data Level",
    )
    group: Optional[Group] = Field(
        None,
        description="A short name for the group of individuals that collected this data",
        title="Group",
    )
    project_name: Optional[str] = Field(
        None,
        description="A name for a set of coordinated activities intended to achieve one or more objectives",
        title="Project Name",
    )
    project_id: Optional[str] = Field(
        None,
        description="A database or other identifier for a project",
        title="Project ID",
    )
    restrictions: Optional[str] = Field(
        None,
        description="Detail any restrictions on publishing or sharing these data",
        title="Restrictions",
    )
    modality: str = Field(
        ...,
        regex=DataRegex.NO_UNDERSCORES.value,
        description="A short name for the specific manner, characteristic, pattern of application, or the employment of any technology or formal procedure to generate data for a study",
        title="Modality",
    )
    subject_id: str = Field(
        ...,
        regex=DataRegex.NO_UNDERSCORES.value,
        description="Unique identifier for the subject of data acquisition",
    )

    def __init__(self, label, **kwargs):
        """Construct a generic data description"""
        name = build_data_name(
            label=label,
            creation_date=kwargs["creation_date"],
            creation_time=kwargs["creation_time"],
        )

        super().__init__(name=name, **kwargs)

    @classmethod
    def parse_name(cls, name):
        """Decompose a DataDescription name string into component parts"""
        m = re.match(f"{DataRegex.DATA.value}", name)

        if m is None:
            raise ValueError(f"name({name}) does not match pattern")

        creation_date, creation_time = datetime_from_name_string(
            m.group("c_date"), m.group("c_time")
        )

        return dict(
            label=m.group("label"),
            creation_date=creation_date,
            creation_time=creation_time,
        )

    @classmethod
    def from_name(cls, name, **kwargs):
        """construct a DataDescription from a name string"""
        d = cls.parse_name(name)

        return cls(**d, **kwargs)


class DerivedDataDescription(DataDescription):
    """A logical collection of data files derived via processing"""

    input_data_name: str
    data_level: DataLevel = Field(
        DataLevel.DERIVED_DATA,
        description="level of processing that data has undergone",
        title="Data Level",
        const=True,
    )

    def __init__(self, process_name, **kwargs):
        """Construct a derived data description"""
        input_data_name = kwargs["input_data_name"]
        super().__init__(label=f"{input_data_name}_{process_name}", **kwargs)

    @classmethod
    def parse_name(cls, name):
        """decompose DerivedDataDescription name into parts"""

        # look for input data name
        m = re.match(f"{DataRegex.DERIVED_DATA.value}", name)

        if m is None:
            raise ValueError(f"name({name}) does not match pattern")

        creation_date, creation_time = datetime_from_name_string(
            m.group("c_date"), m.group("c_time")
        )

        return dict(
            process_name=m.group("process_name"),
            creation_date=creation_date,
            creation_time=creation_time,
            input_data_name=m.group("input"),
        )

    @classmethod
    def from_name(cls, name, **kwargs):
        """build DerivedDataDescription from a name"""

        d = cls.parse_name(name)

        return cls(**d, **kwargs)

    @classmethod
    def from_data_description(cls, input_data, process_name, **kwargs):
        """Build a DerivedDataDescription from an input DataDescription"""

        if input_data.data_level == DataLevel.DERIVED_DATA:
            name_len = len(input_data.input_data_name) + 1
            input_data_name = input_data.name[name_len:]
        else:
            input_data_name = input_data.name

        return cls(
            input_data_name=input_data_name,
            process_name=process_name,
            subject_id=input_data.subject_id,
            modality=input_data.modality,
            **kwargs,
        )


class RawDataDescription(DataDescription):
    """A logical collection of data files as acquired from a rig or instrument"""

    data_level: DataLevel = Field(
        DataLevel.RAW_DATA,
        description="level of processing that data has undergone",
        title="Data Level",
        const=True,
    )

    def __init__(self, **kwargs):
        """Construct a raw data description"""
        modality = kwargs["modality"]
        subject_id = kwargs["subject_id"]
        super().__init__(label=f"{modality}_{subject_id}", **kwargs)

    @classmethod
    def parse_name(cls, name):
        """Decompose raw data description name into component parts"""

        m = re.match(f"{DataRegex.RAW_DATA.value}", name)

        if m is None:
            raise ValueError(f"name({name}) does not match pattern")

        creation_date, creation_time = datetime_from_name_string(
            m.group("c_date"), m.group("c_time")
        )

        return dict(
            modality=m.group("modality"),
            subject_id=m.group("subject_id"),
            creation_date=creation_date,
            creation_time=creation_time,
        )

    @classmethod
    def from_name(cls, name, **kwargs):
        """construct from a name string"""

        d = cls.parse_name(name)

        return cls(**d, **kwargs)



""" example processing """

import datetime
from aind_data_schema import Processing
from aind_data_schema.processing import DataProcess

t = datetime.datetime(2022, 11, 22, 8, 43, 00)

p = Processing(
    pipeline_url="https://url/for/pipeline",
    pipeline_version="0.1.0",
    data_processes=[
        DataProcess(
            name="Image tile fusing",
            version="0.0.1",
            start_date_time=t,
            end_date_time=t,
            input_location="/path/to/inputs",
            output_location="/path/to/outputs",
            code_url="https://github.com/abcd",
            parameters={"size": 7},
        ),
        DataProcess(
            name="File format conversion",
            version="0.0.1",
            start_date_time=t,
            end_date_time=t,
            input_location="/path/to/inputs",
            output_location="/path/to/outputs",
            code_url="https://github.com/asdf",
            parameters={"u": 7, "z": True},
        ),
        DataProcess(
            name="Image destriping",
            version="0.2.1",
            start_date_time=t,
            end_date_time=t,
            input_location="/path/to/input",
            output_location="/path/to/output",
            code_url="https://github.com/fdsa",
            parameters={"a": 2, "b": -2},
        ),
    ],
)

with open("processing.json", "w") as f:
    f.write(p.json(indent=3))



"""Configuration file for the Sphinx documentation builder."""
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path Setup --------------------------------------------------------------
import os
import pathlib
import sys

from pygit2 import Repository

sys.path.insert(0, os.path.abspath("../../src"))
# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information
head = Repository(pathlib.Path.cwd())

project = str(head).split("/")[-3]
copyright = "2022, Allen Institute of Neural Dynamics"
author = "Allen Institute of Neural Dynamics"
release = "0.1"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.duration",
    "sphinx.ext.doctest",
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
]
templates_path = ["_templates"]
exclude_patterns = []

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "alabaster"
html_static_path = ["_static"]

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
html_show_sphinx = False

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
html_show_copyright = False



""" Schemas for Ophys Rigs"""

from __future__ import annotations

from datetime import date, datetime
from enum import Enum
from typing import List, Optional

from pydantic import Field
from ..base import AindCoreModel

from ..device import Device


class CameraName(Enum):
    """Camera name"""

    BODY_CAMERA = "Body Camera"
    EYE_CAMERA = "Eye Camera"
    FACE_CAMERA = "Face Camera"


class Camera(Device):
    """Description of an ophys camera"""

    name: CameraName = Field(..., title="Camera Name")
    position_x: float = Field(..., title="Position X")
    position_y: float = Field(..., title="Position Y")
    position_z: float = Field(..., title="Position Z")
    angle_pitch: float = Field(..., title="Angle pitch (deg)")
    angle_yaw: float = Field(..., title="Angle yaw (deg)")
    angle_roll: float = Field(..., title="Angle roll (deg)")
    recording_software: Optional[str] = Field(None, title="Recording software")
    recording_software_version: Optional[str] = Field(
        None, title="Recording software version"
    )


class CameraType(Enum):
    """Camera type name"""

    CAMERA = "Camera"
    PMT = "PMT"
    OTHER = "other"


class DataInterface(Enum):
    """Data interface name"""

    USB = "USB"
    CAMERALINK = "CameraLink"
    COAX = "Coax"
    OTHER = "other"


class Cooling(Enum):
    """Cooling medium name"""

    AIR = "air"
    WATER = "water"


class Immersion(Enum):
    """Immersion medium name"""

    AIR = "air"
    WATER = "water"
    OIL = "oil"


class Detector(Device):
    """Description of a generic detector"""

    name: Optional[str] = Field(
        None,
        description="Brief name to identify detector to match with session information",
        title="Name",
    )
    type: CameraType = Field(..., title="Camera Type")
    data_interface: DataInterface = Field(..., title="Data interface")
    cooling: Cooling = Field(..., title="Cooling")
    immersion: Optional[Immersion] = Field(None, title="Immersion")


class FilterType(Enum):
    """Filter type name"""

    LONG_PASS = "Long pass"
    BAND_PASS = "Band pass"


class FilterSize(Enum):
    """Filter size value"""

    FILTER_SIZE_25 = 25
    FILTER_SIZE_32 = 32


class Filter(Device):
    """Description of a filter"""

    type: FilterType = Field(..., title="Filter Type")
    size: Optional[FilterSize] = Field(None, title="Size (mm)")
    cut_off_frequency: Optional[int] = Field(None, title="Cut off frequency")
    cut_on_frequency: Optional[int] = Field(None, title="Cut on frequency")
    description: Optional[str] = Field(
        None, description="Where/how filter is being used", title="Description"
    )


class LaserName(Enum):
    """Laser name"""

    LASER_A = "Laser A"
    LASER_B = "Laser B"
    LASER_C = "Laser C"
    LASER_D = "Laser D"
    LASER_E = "Laser E"


class Coupling(Enum):
    """Coupling type name"""

    FREE_SPACE = "Free-space"
    SMF = "SMF"
    MMF = "MMF"
    OTHER = "other"


class Laser(Device):
    """Description of a laser"""

    name: LaserName = Field(..., title="Laser Name")
    item_number: Optional[str] = Field(None, title="Item number")
    wavelength: int = Field(..., title="Wavelength (nm)")
    maximum_power: float = Field(..., title="Maximum power (mW)")
    coupling: Optional[Coupling] = Field(None, title="Coupling")
    coupling_efficiency: Optional[float] = Field(
        None, title="Coupling efficiency (percent)"
    )
    calibration_data: Optional[str] = Field(
        None, description="path to calibration data", title="Calibration data"
    )
    calibration_date: Optional[datetime] = Field(
        None, title="Calibration date"
    )


class LensSize(Enum):
    """Lens size value"""

    LENS_SIZE_1 = 1
    LENS_SIZE_2 = 2


class Lens(Device):
    """Description of a lens"""

    focal_length: Optional[float] = Field(None, title="Focal length (mm)")
    size: Optional[LensSize] = Field(None, title="Size (inches)")
    optimized_wavelength_range: Optional[str] = Field(
        None, title="Optimized wavelength range (nm)"
    )
    notes: Optional[str] = Field(None, title="Notes")


class PatchName(Enum):
    """Patch name"""

    PATCH_CORD_A = "Patch Cord A"
    PATCH_CORD_B = "Patch Cord B"
    PATCH_CORD_C = "Patch Cord C"


class Patch(Device):
    """Description of a patch"""

    name: PatchName = Field(..., title="Patch Name")
    core_diameter: float = Field(..., title="Core diameter (um)")
    numerical_aperture: float = Field(..., title="Numerical aperture")
    photobleaching_date: Optional[date] = Field(
        None, title="Photobleaching date"
    )


class OphysRig(AindCoreModel):
    """Description of an optical physiology rig"""

    schema_version: str = Field(
        "0.2.0",
        description="schema version",
        title="Schema Version",
        const=True,
    )
    rig_id: str = Field(
        ..., description="room number_stim apparatus_version", title="Rig ID"
    )
    rig_location: Optional[str] = Field(None, title="Rig location")
    temperature_control: Optional[bool] = Field(
        None, title="Temperature control"
    )
    humidity_control: Optional[bool] = Field(None, title="Humidity control")
    vibration_control: Optional[bool] = Field(None, title="Vibration control")
    patch_cords: List[Patch] = Field(
        ..., title="Patch cords", unique_items=True
    )
    cameras: Optional[List[Camera]] = Field(
        None, title="Cameras", unique_items=True
    )
    lasers: List[Laser] = Field(..., title="Lasers", unique_items=True)
    detectors: Optional[List[Detector]] = Field(
        None, title="Detectors", unique_items=True
    )
    filters: Optional[List[Filter]] = Field(
        None, title="Filters", unique_items=True
    )
    lenses: Optional[List[Lens]] = Field(
        None, title="Lenses", unique_items=True
    )
    additional_devices: Optional[List[Device]] = Field(
        None, title="Additional devices", unique_items=True
    )
    light_path_diagram: Optional[str] = Field(
        None,
        description="Path to diagram of the light path.",
        title="Light path diagram",
    )
    notes: Optional[str] = Field(None, title="Notes")



""" example SmartSPIM instrument """

from aind_data_schema.imaging import instrument

inst = instrument.Instrument(
    instrument_id="SmartSPIM2-1",
    type="smartSPIM",
    manufacturer="LifeCanvas",
    location="440 Westlake",
    objectives=[
        instrument.Objective(
            numerical_aperture=0.2,
            magnification=3.6,
            manufacturer="LifeCanvas",
            immersion="multi",
            notes="Thorlabs TL4X-SAP with LifeCanvas dipping cap and correction optics.",
            serial_number="Unknown-1",
        ),
        instrument.Objective(
            numerical_aperture=0.12,
            magnification=1.625,
            manufacturer="LifeCanvas",
            immersion="multi",
            notes="Thorlabs TL2X-SAP with LifeCanvas dipping cap and correction optics.",
            serial_number="Unknown-2",
        ),
    ],
    detectors=[
        instrument.Detector(
            type="Camera",
            data_interface="USB",
            cooling="air",
            manufacturer="Hamamatsu",
            model="C14440-20UP",
            serial_number="001284",
        ),
    ],
    light_sources=[
        instrument.Lightsource(
            type="laser",
            coupling="SMF",
            wavelength=445,
            max_power=200,
            serial_number="VL08223M03",
            manufacturer="Vortran",
        ),
        instrument.Lightsource(
            type="laser",
            coupling="SMF",
            wavelength=488,
            max_power=150,
            serial_number="VL08223M03",
            manufacturer="Vortran",
        ),
        instrument.Lightsource(
            type="laser",
            coupling="SMF",
            wavelength=561,
            max_power=150,
            serial_number="VL08223M03",
            manufacturer="Vortran",
        ),
        instrument.Lightsource(
            type="laser",
            coupling="SMF",
            wavelength=594,
            max_power=100,
            serial_number="VL08223M03",
            manufacturer="Vortran",
        ),
        instrument.Lightsource(
            type="laser",
            coupling="SMF",
            wavelength=639,
            max_power=160,
            serial_number="VL08223M03",
            manufacturer="Vortran",
        ),
        instrument.Lightsource(
            type="laser",
            coupling="SMF",
            wavelength=690,
            max_power=160,
            serial_number="VL08223M03",
            manufacturer="Vortran",
        ),
    ],
    fluorescence_filters=[
        instrument.Filter(
            type="Band pass",
            manufacturer="Semrock",
            diameter=25,
            thickness=2.0,
            model="FF01-469/35-25",
            filter_wheel_index=0,
            serial_number="Unknown-0",
        ),
        instrument.Filter(
            type="Band pass",
            manufacturer="Semrock",
            diameter=25,
            thickness=2.0,
            model="FF01-525/45-25",
            filter_wheel_index=1,
            serial_number="Unknown-1",
        ),
        instrument.Filter(
            type="Band pass",
            manufacturer="Semrock",
            diameter=25,
            thickness=2.0,
            model="FF01-593/40-25",
            filter_wheel_index=2,
            serial_number="Unknown-2",
        ),
        instrument.Filter(
            type="Band pass",
            manufacturer="Semrock",
            diameter=25,
            thickness=2.0,
            model="FF01-624/40-25",
            filter_wheel_index=3,
            serial_number="Unknown-3",
        ),
        instrument.Filter(
            type="Band pass",
            manufacturer="Chroma",
            diameter=25,
            thickness=2.0,
            model="ET667/30m",
            filter_wheel_index=4,
            serial_number="Unknown-4",
        ),
        instrument.Filter(
            type="Long pass",
            manufacturer="Thorlabs",
            diameter=25,
            thickness=2.0,
            model="FELH0700",
            filter_wheel_index=5,
            serial_number="Unknown-5",
        ),
    ],
    motorized_stages=[
        instrument.MotorizedStage(
            model="LS-100",
            manufacturer="Applied Scientific Instrumentation",
            serial_number="Unknown-1",
            travel=100,
            notes="Focus stage",
        ),
        instrument.MotorizedStage(
            model="L12-20F-4",
            manufacturer="IR Robot Co",
            serial_number="Unknown-5",
            travel=41,
            notes="Cylindrical lens #1",
        ),
        instrument.MotorizedStage(
            model="L12-20F-4",
            manufacturer="IR Robot Co",
            serial_number="Unknown-6",
            travel=41,
            notes="Cylindrical lens #2",
        ),
        instrument.MotorizedStage(
            model="L12-20F-4",
            manufacturer="IR Robot Co",
            serial_number="Unknown-7",
            travel=41,
            notes="Cylindrical lens #3",
        ),
        instrument.MotorizedStage(
            model="L12-20F-4",
            manufacturer="IR Robot Co",
            serial_number="Unknown-8",
            travel=41,
            notes="Cylindrical lens #4",
        ),
    ],
    scanning_stages=[
        instrument.ScanningStage(
            model="LS-50",
            manufacturer="Applied Scientific Instrumentation",
            serial_number="Unknown-2",
            stage_axis_direction="Detection axis",
            stage_axis_name="Z",
            travel=50,
            notes="Sample stage Z",
        ),
        instrument.ScanningStage(
            model="LS-50",
            manufacturer="Applied Scientific Instrumentation",
            serial_number="Unknown-3",
            stage_axis_direction="Illumination axis",
            stage_axis_name="X",
            travel=50,
            notes="Sample stage X",
        ),
        instrument.ScanningStage(
            model="LS-50",
            manufacturer="Applied Scientific Instrumentation",
            serial_number="Unknown-4",
            stage_axis_direction="Perpendicular axis",
            stage_axis_name="Y",
            travel=50,
            notes="Sample stage Y",
        ),
    ],
    optical_tables=[
        instrument.OpticalTable(
            length=36,
            width=48,
            vibration_control=False,
            model="VIS2424-IG2-125A",
            manufacturer="MKS Newport",
            serial_number="Unknown",
        )
    ],
    com_ports=[
        instrument.Com(hardware_name="Laser Launch", com_port="COM4"),
        instrument.Com(
            hardware_name="ASI Tiger",
            com_port="COM3",
        ),
        instrument.Com(
            hardware_name="MightyZap",
            com_port="COM9",
        ),
    ],
    humidity_control=False,
    temperature_control=False,
)

with open("aind_smartspim_instrument.json", "w") as f:
    f.write(inst.json(indent=3))



""" example subject """

from aind_data_schema import Subject
import datetime

t = datetime.datetime(2022, 11, 22, 8, 43, 00)

s = Subject(
    species="Mus musculus",
    subject_id="12345",
    sex="Male",
    date_of_birth=t.date(),
    genotype="Emx1-IRES-Cre;Camk2a-tTA;Ai93(TITL-GCaMP6f)",
    home_cage_enrichment="other",
    background_strain="C57BL/6J",
)

with open("subject.json", "w") as f:
    f.write(s.json(indent=3))



""" ephys session description and related objects """

from __future__ import annotations

from datetime import datetime
from enum import Enum
from typing import List, Optional

from pydantic import Field
from ..base import AindCoreModel, AindModel


class SessionType(Enum):
    """Session type name"""

    TEST = "Test"
    OPTO = "Optotagging"
    VISUAL_ORIENTATION = "Visual Orientation"


class ExpectedDataStream(Enum):
    """Names of data streams to expect in recording"""

    NEUROPIXELS_PROBES = "Neuropixels probes"
    BODY_CAMERA = "Body camera"
    FACE_CAMERA = "Face camera"
    EYE_CAMERA = "Eye camera"
    BONSAI_FILE = "Bonsai file"
    HARP_FILE = "Harp bin file"
    OTHER = "Other"


class CcfVersion(Enum):
    """CCF version"""

    CCFv3 = "CCFv3"


class Coordinates3d(AindModel):
    """Description of 3d coordinates in mm"""

    x: float = Field(..., title="X (mm)", units="mm")
    y: float = Field(..., title="Y (mm)", units="mm")
    z: float = Field(..., title="Z (mm)", units="mm")


class CcfCoords(AindModel):
    """Coordinates in CCF template space"""

    ml: float = Field(..., title="ML (um)", units="um")
    ap: float = Field(..., title="AP (um)", units="um")
    dv: float = Field(..., title="DV (um)", units="um")
    ccf_version: CcfVersion = Field(CcfVersion.CCFv3, title="CCF version")


class LaserModule(AindModel):
    """Description of a laser"""

    name: str = Field(..., title="Name")
    wavelength: int = Field(..., title="Wavelength (nm)", units="nm")
    power: float = Field(..., title="Power (mW)", units="mW")
    targeted_structure: str = Field(..., title="Targeted structure")
    targeted_ccf_coordinates: CcfCoords = Field(
        ...,
        title="Targeted CCF coordinates",
    )
    manipulator_coordinates: Coordinates3d = Field(
        ...,
        title="Manipulator coordinates",
    )


class EphysProbe(AindModel):
    """Description of an ephys probe"""

    name: str = Field(..., title="Name")
    tip_targeted_structure: str
    other_targeted_structures: Optional[List[str]] = None
    targeted_ccf_coordinates: Optional[CcfCoords] = Field(
        None,
        title="Targeted CCF coordinates",
    )
    manipulator_coordinates: Coordinates3d = Field(
        ...,
        title="Manipulator coordinates",
    )


class Stream(AindModel):
    """Stream of data with a start and stop time"""

    stream_start_time: datetime = Field(..., title="Stream start time")
    stream_stop_time: datetime = Field(..., title="Stream stop time")
    probes: List[EphysProbe] = Field(..., title="Probes", unique_items=True)
    lasers: List[LaserModule] = Field(..., title="Lasers", unique_items=True)


class EphysSession(AindCoreModel):
    """Description of an ephys recording session"""

    schema_version: str = Field(
        "0.2.1", description="schema version", title="Version", const=True
    )
    experimenter_full_name: str = Field(
        ...,
        description="First and last name of the experimenter.",
        title="Experimenter full name",
    )
    session_start_time: datetime = Field(..., title="Session start time")
    session_end_time: datetime = Field(..., title="Session end time")
    subject_id: int = Field(..., title="Subject ID")
    session_type: SessionType = Field(..., title="Session type")
    session_description: Optional[str] = Field(
        None, title="Session description"
    )
    stimulus_protocol_id: Optional[str] = Field(
        None, title="Stimulus protocol ID"
    )
    iacuc_protocol: Optional[str] = Field(None, title="IACUC protocol")
    rig_id: str = Field(..., title="Rig ID")
    expected_data_streams: Optional[List[ExpectedDataStream]] = None
    probe_streams: List[Stream] = Field(
        ..., title="Probe streams", unique_items=True
    )
    ccf_coordinate_transform: Optional[str] = Field(
        None,
        description="Path to file that details the CCF-to-lab coordinate transform.",
        title="CCF coordinate transform",
    )
    notes: Optional[str] = None



""" schema for various Procedures """

from __future__ import annotations

from datetime import date, time
from enum import Enum
from typing import List, Optional

from pydantic import Field
from .base import AindCoreModel, AindModel


class ProtectiveMaterial(Enum):
    """Name of material applied to craniotomy"""

    DURAGEL = "Duragel"
    SORTA_CLEAR = "SORTA-clear"
    KWIK_CAST = "Kwik-Cast"
    OTHER = "Other - see notes"


class Anaesthetic(AindModel):
    """Description of an anaestheic"""

    type: str = Field(..., title="Type")
    duration: float = Field(..., title="Duration (hours)", units="hours")
    level: float = Field(
        ..., title="Level (percent)", units="percent", ge=1, le=5
    )


class Procedure(AindModel):
    """Description of surgical or other procedure performed on a subject"""

    type: Optional[str] = Field(
        None, description="Procedure type", title="Procedure Type"
    )
    start_date: date = Field(..., title="Start date")
    end_date: date = Field(..., title="End date")
    experimenter_full_name: str = Field(
        ...,
        description="First and last name of the experimenter.",
        title="Experimenter full name",
    )
    protocol_id: str = Field(
        ..., title="Protocol ID", description="DOI for protocols.io"
    )
    iacuc_protocol: Optional[str] = Field(None, title="IACUC protocol")
    animal_weight: Optional[float] = Field(
        None,
        title="Animal weight (g)",
        description="Animal weight before procedure",
        units="g",
    )
    anaesthesia: Optional[Anaesthetic] = Field(None, title="Anaesthesia")
    notes: Optional[str] = Field(None, title="Notes")


class Craniotomy(Procedure):
    """Description of craniotomy procedure"""

    craniotomy_coordinates_ml: float = Field(
        ..., title="Craniotomy coordinate ML (mm)", units="mm"
    )
    craniotomy_coordinates_ap: float = Field(
        ..., title="Craniotomy coordinates AP (mm)", units="mm"
    )
    craniotomy_size: float = Field(
        ..., title="Craniotomy size (mm)", units="mm"
    )
    implant_part_number: Optional[str] = Field(
        None, title="Implant part number"
    )
    dura_removed: Optional[bool] = Field(None, title="Dura removed")
    protective_material: Optional[ProtectiveMaterial] = Field(
        None, title="Protective material"
    )
    workstation_id: Optional[str] = Field(None, title="Workstation ID")


class HeadframeMaterial(Enum):
    """Headframe material name"""

    TITANIUM = "Titanium"
    STEEL = "Steel"


class Headframe(Procedure):
    """Description of headframe procedure"""

    headframe_part_number: str = Field(..., title="Headframe part number")
    headframe_material: HeadframeMaterial = Field(
        ..., title="Headframe material"
    )
    well_part_number: Optional[str] = Field(None, title="Well part number")
    well_type: Optional[str] = Field(None, title="Well type")


class Side(Enum):
    """Side of animal"""

    LEFT = "Left"
    RIGHT = "Right"


class VirusPrepType(Enum):
    """Type of virus preparation"""

    CRUDE = "Crude"
    PURIFIED = "Purified"


class InjectionMaterial(AindModel):
    """Description of injection material"""

    name: str = Field(..., title="Name")
    material_id: Optional[str] = Field(None, title="Material ID")
    full_genome_name: Optional[str] = Field(
        None,
        title="Full genome name",
        description="Full genome for virus construct",
    )
    plasmid_name: Optional[str] = Field(
        None,
        title="Plasmid name",
        description="Short name used to reference the plasmid",
    )
    genome_copy: Optional[float] = Field(None, title="Genome copy")
    titer: Optional[float] = Field(None, title="Titer (gc/mL", units="gc/mL")
    prep_lot_number: Optional[str] = Field(
        None, title="Preparation lot number"
    )
    prep_date: Optional[date] = Field(
        None,
        title="Preparation lot date",
        description="Date this prep lot was titered",
    )
    prep_type: Optional[VirusPrepType] = Field(None, title="Viral prep type")


class Injection(Procedure):
    """Description of an injection procedure"""

    injection_materials: List[InjectionMaterial] = Field(
        None, title="Injection material", unique_items=True
    )
    injection_duration: time = Field(..., title="Injection duration")
    recovery_time: float = Field(..., title="Recovery time (min)", units="min")
    workstation_id: Optional[str] = Field(None, title="Workstation ID")
    instrument_id: Optional[str] = Field(None, title="Instrument ID")


class RetroOrbitalInjection(Injection):
    """Description of a retro-orbital injection procedure"""

    injection_type: str = Field(
        "Retro-orbital", title="Injection type", const=True
    )
    injection_volume: float = Field(
        ..., title="Injection volume (nL)", units="nL"
    )
    injection_eye: Side = Field(..., title="Injection eye")


class BrainInjection(Injection):
    """Description of a brain injection procedure"""

    injection_hemisphere: Optional[Side] = Field(
        None, title="Injection hemisphere"
    )
    injection_coordinate_ml: float = Field(
        ..., title="Injection coordinate ML (mm)"
    )
    injection_coordinate_ap: float = Field(
        ..., title="Injection coordinate AP (mm)"
    )
    injection_coordinate_depth: float = Field(
        ..., title="Injection coodinate depth (mm)"
    )
    injection_angle: float = Field(
        ..., title="Injection angle (deg)", units="deg"
    )


class NanojectInjection(BrainInjection):
    """Description of a nanoject injection procedure"""

    injection_type: str = Field("Nanoject", title="Injection type", const=True)
    injection_volume: float = Field(
        ..., title="Injection volume (nL)", units="nL"
    )


class IontophoresisInjection(BrainInjection):
    """Description of an iotophoresis injection procedure"""

    injection_type: str = Field(
        "Iontophoresis", title="Injection type", const=True
    )
    injection_current: float = Field(
        ..., title="Injection current (μA)", units="μA"
    )
    alternating_current: str = Field(..., title="Alternating current")


class IntraCerebellarVentricleInjection(BrainInjection):
    """Description of an interacerebellar ventricle injection"""

    injection_type: str = Field("ICV", title="Injection type", const=True)
    injection_volume: float = Field(
        ..., title="Injection volume (nL)", units="nL"
    )


class IntraCisternalMagnaInjection(BrainInjection):
    """Description of an interacisternal magna injection"""

    injection_type: str = Field("ICM", title="Injection type", const=True)
    injection_volume: float = Field(
        ..., title="Injection volume (nL)", units="nL"
    )


class MriScanSequence(Enum):
    """MRI scan sequence"""

    RARE = "RARE"


class ScannerLocation(Enum):
    """location of scanner"""

    UW_SLU = "UW SLU"
    FRED_HUTCH = "Fred Hutch"


class MagneticStrength(Enum):
    """Strength of magnet"""

    MRI_7T = 7
    MRI_14T = 14


class MriScan(Procedure):
    """Description of an MRI scan"""

    scan_sequence: MriScanSequence = Field(..., title="Scan sequence")
    scanner_location: Optional[ScannerLocation] = Field(
        None, title="Scanner location"
    )
    magnetic_strength: Optional[MagneticStrength] = Field(
        None, title="Magnetic strength (T)", units="T"
    )
    resolution: float = Field(..., title="Resolution")
    protocol_id: str = Field(..., title="Protocol ID")


class TissuePrepName(Enum):
    """Tissue preparation type name"""

    PERFUSION = "Perfusion"
    FIXATION = "Fixation"
    DOUBLE_DELIPIDATION = "Double delipidation"
    DCM_DELIPIDATION = "DCM delipidation"
    IMMUNOSTAINING = "Immunostaining"
    GELATION = "Gelation"


class TissuePrep(Procedure):
    """Description of a tissue preparation procedure"""

    name: TissuePrepName = Field(..., title="Name")
    specimen_id: List[str] = Field(..., title="Specimen ID")


class TrainingProtocol(AindModel):
    """Description of an animal training protocol"""

    protocol_id: str = Field(..., title="Training protocol ID")
    training_protocol_start_date: date = Field(
        ..., title="Training protocol start date"
    )
    training_protocol_end_date: Optional[date] = Field(
        None, title="Training protocol end date"
    )
    notes: Optional[str] = Field(None, title="Notes")


class ProbeName(Enum):
    """Probe name"""

    PROBE_A = "Probe A"
    PROBE_B = "Probe B"
    PROBE_C = "Probe C"


class FerruleMaterial(Enum):
    """Probe ferrule material type name"""

    CERAMIC = "Ceramic"
    STAINLESS_STEEL = "Stainless steel"


class OphysProbe(AindModel):
    """Description of an ophys probe"""

    name: ProbeName = Field(..., title="Name")
    manufacturer: str = Field(..., title="Manufacturer")
    part_number: str = Field(..., title="Part number")
    core_diameter: float = Field(..., title="Core diameter (μm)", units="μm")
    numerical_aperture: float = Field(..., title="Numerical aperture")
    ferrule_material: Optional[FerruleMaterial] = Field(
        None, title="Ferrule material"
    )
    targeted_structure: str = Field(..., title="Targeted structure")
    stereotactic_coordinate_ap: float = Field(
        ..., title="Stereotactic coordinate A/P (mm)", units="mm"
    )
    stereotactic_coordinate_ml: float = Field(
        ..., title="Stereotactic coodinate M/L (mm)", units="mm"
    )
    stereotactic_coordinate_dv: float = Field(
        ..., title="Stereotactic coordinate D/V (mm)", units="mm"
    )
    angle: float = Field(..., title="Angle (deg)", units="deg")
    notes: Optional[str] = Field(None, title="Notes")


class FiberImplant(Procedure):
    """Description of an implant procedure"""

    probes: List[OphysProbe] = Field(
        ..., title="Ophys Probes", unique_items=True
    )


class WaterRestriction(AindModel):
    """Description of a water restriction procedure"""

    protocol_id: Optional[str] = Field(
        None, title="Water restriction protocol number"
    )
    baseline_weight: float = Field(
        ...,
        title="Baseline weight (g)",
        description="Weight at start of water restriction",
    )
    start_date: date = Field(..., title="Water restriction start date")
    end_date: date = Field(..., title="Water restriction end date")


class Procedures(AindCoreModel):
    """Description of all procedures performed on a subject"""

    schema_version: str = Field(
        "0.4.4", description="schema version", title="Version", const=True
    )
    subject_id: str = Field(
        ...,
        description="Unique identifier for the subject. If this is not a Allen LAS ID, indicate this in the Notes.",
        title="Subject ID",
    )
    headframes: Optional[List[Headframe]] = Field(
        None, title="Headframes", unique_items=True
    )
    craniotomies: Optional[List[Craniotomy]] = Field(
        None, title="Craniotomies", unique_items=True
    )
    mri_scans: Optional[List[MriScan]] = Field(
        None, title="MRI scans", unique_items=True
    )
    injections: Optional[List[Injection]] = Field(
        None, title="Injections", unique_items=True
    )
    fiber_implants: Optional[List[FiberImplant]] = Field(
        None, title="Fiber implants", unique_items=True
    )
    water_restriction: Optional[WaterRestriction] = Field(
        None, title="Water restriction"
    )
    training_protocols: Optional[List[TrainingProtocol]] = Field(
        None, title="Training protocols", unique_items=True
    )
    tissue_preparations: Optional[List[TissuePrep]] = Field(
        None, title="Tissue preparations", unique_items=True
    )
    other_procedures: Optional[List[Procedure]] = Field(
        None, title="Other procedures", unique_items=True
    )
    notes: Optional[str] = Field(None, title="Notes")
